#include <stdint.h>
#include <stdio.h>
#include <string.h>

#define TLV_HEADER_LENGTH 3u

typedef enum __attribute__((packed)) Protocol_FrameType_e
{
    /* Anchor -> PC*/
    Protocol_Frame_ToF                   = 0x01,
    Protocol_Frame_Serial                = 0x02,
    Protocol_Frame_AdminDistributionData = 0x03,
    Protocol_Frame_CompressedData        = 0x04,
    Protocol_Frame_AccelerometerData     = 0x05,
    Protocol_Frame_GyroscopeData         = 0x06,
    Protocol_Frame_MagnetometerData      = 0x07,
    Protocol_Frame_ImpactData            = 0x08,
    Protocol_Frame_Sync_Blink            = 201,
    Protocol_Frame_Sync_Shutdown         = 140,
    Protocol_Frame_SetUWBConfig          = 139,
    
} Protocol_FrameType_t;

typedef struct __attribute__((packed)) Protocol_Frame_s
{
    uint8_t type;
    uint16_t length;
    uint8_t value[];

} Protocol_Frame_t;



typedef struct __attribute__((packed)) ToF_TagImpactReport_s
{
    uint8_t did[3];
    uint8_t hitCycleOffset;
    uint8_t motionByte;

} ToF_TagImpactReport_t;

typedef struct __attribute__((packed)) ToF_Tag_s
{
    uint8_t did[3];
    struct
    {
        uint8_t receivedByTag : 4;    /** bits 0:3 */
        uint8_t receivedByAnchor : 4; /** bits 4:7 */
    } rssi;

    uint16_t tof[];
} ToF_Tag_t;

typedef struct __attribute__((packed)) ToF_TagList_s
{
    struct
    {
        uint8_t numberOfToFs : 6; /** bits 0:5 */
        uint8_t precision : 2;    /** bits 6:7 */
    };

    uint8_t tagCount;
    uint8_t tagData[];

} ToF_TagList_t;

typedef struct __attribute__((packed)) Protocol_ToFReport_s
{
    uint8_t did[3];
    uint16_t cycleId;

    struct
    {
        uint8_t did[3];
        struct
        {
            uint8_t receivedByTag : 4;    /** bits 0:3 */
            uint8_t receivedByAnchor : 4; /** bits 4:7 */
        } rssi;

        uint16_t tof[4];
    } a2a;

    struct
    {
        uint8_t hitCycleOffset;
        uint8_t motionByte;
        uint8_t tagImpactReportCount;
    } imu;

    uint8_t tagListsCount;
    uint8_t tagListsData[];

} Protocol_ToFReport_t;

typedef struct __attribute__((packed)) Protocol_Sensors_s
{
    uint8_t did[3];
    uint16_t cycleId;
    uint8_t slotId;
    uint8_t data[];
} Protocol_Sensors_t;

typedef struct __attribute__((packed)) Protocol_AdminDistributionInfo_s
{
    uint8_t did[3];
    uint16_t cycleId;

    /* Constants */

    uint8_t apiVersion;
    uint8_t fwVersion;
    uint8_t hwid;
    uint8_t MAC[6];

    /* Config */
    struct
    {
        uint8_t deviceOffset;
        uint8_t frequency;
        uint8_t seat;
    } cycle;

    uint8_t serialNumber[4];
    char name[9];
    char group[3];
    uint8_t license[2];

    struct
    {
        uint8_t txPowerPreset;
    } bt;

    struct
    {
        uint8_t ch3AntennaDelay;
        uint8_t ch5AntennaDelay;
        uint8_t txPowerPreset;
    } uwb;

    struct
    {
        uint8_t boardConfig;
        uint8_t pressureCallibration;

        struct
        {
            uint8_t x;
            uint8_t y;
            uint8_t z;
        } magnetoBias;

        struct
        {
            uint8_t x;
            uint8_t y;
            uint8_t z;
        } magnetoScalar;

        struct
        {
            uint8_t x;
            uint8_t y;
            uint8_t z;
        } accelerometerBias;

    } sensorsConfig;

    /* Dynamic sensor data */
    struct
    {
        uint8_t vbat;
        uint8_t motionByte;
        uint8_t hit;
        uint8_t pressure;
        uint8_t heartRate;
        uint8_t mox;
    } sensorsData;

    uint8_t valid[7];

} Protocol_AdminDistributionInfo_t;

typedef struct __attribute__((packed)) SM_Protocol_CommandShutdown_s
{
    struct __attribute__((packed))
    {
        uint8_t did[3];
    } device[4];

} SM_Protocol_CommandShutdown_t;

typedef struct __attribute__((packed)) SM_Protocol_CommandBlink_s
{
    struct __attribute__((packed))
    {
        uint8_t did[3];
        struct
        {
            uint8_t pattern : 6;    /** bits 0:5 */
            uint8_t brightness : 2; /** bits 6:7 */
        };

        uint16_t cycleId;
    } device[2];

} SM_Protocol_CommandBlink_t;

typedef struct __attribute__((packed)) SM_Protocol_CommandSetSeatBasic_s
{
    struct __attribute__((packed))
    {
        uint8_t did[3];
        uint8_t seat;
    } device[3];

} SM_Protocol_CommandSetSeatBasic_t;

typedef struct __attribute__((packed)) SM_Protocol_CommandSetUWBConfig_s
{
    struct __attribute__((packed))
    {
        uint8_t did[3];
        int8_t ch3;
        int8_t ch5;
        uint8_t txPower;
    } device[2];

} SM_Protocol_CommandSetUWBConfig_t;

typedef struct __attribute__((packed)) Sensors_Motion_s
{
    struct
    {
        uint8_t length : 1;
        uint8_t hitPosition : 7;
    };

    struct
    {
        uint8_t lsb;
        uint8_t msb;
    } impact;

} Sensors_Motion_t;

#define ARRAY_SIZE(type) (sizeof(type) / sizeof(type[0]))

uint32_t bitStitcher_compress(
    const int32_t* const input,
    uint32_t* const output,
    uint8_t length,
    uint8_t bitsPerSample,
    uint8_t initialOffset)
{
    const uint8_t outputBits = sizeof(*output) * 8u;
    const uint32_t valueMask = ~(-(1 << bitsPerSample));

    uint8_t bitsLeft    = outputBits - initialOffset;
    uint32_t* outputPtr = &output[0];
    for (uint8_t inputIdx = 0; inputIdx < length; ++inputIdx)
    {
        const uint32_t value = input[inputIdx] & valueMask;
        if (bitsLeft >= bitsPerSample)
        {
            *outputPtr |= value << (outputBits - bitsLeft);
            bitsLeft -= bitsPerSample;

            if (bitsLeft == 0)
            {
                ++outputPtr;
                bitsLeft = outputBits;
            }
        }
        else
        {
            const uint32_t part1Mask = ~(-(1 << bitsLeft));
            const uint32_t part1     = value & part1Mask;
            const uint32_t part2     = value >> bitsLeft;
            const uint8_t part2Bits  = bitsPerSample - bitsLeft;

            *outputPtr |= part1 << (outputBits - bitsLeft);
            ++outputPtr;
            *outputPtr |= part2;

            bitsLeft = outputBits - part2Bits;
        }
    }

    return (uint32_t)((uintptr_t)outputPtr - (uintptr_t)output)
           + (sizeof(*output) - (bitsLeft / 8u));
}

static void bitStitcher_helper_print(uint32_t* buffer, uint8_t length)
{
    printf("WORDS\n");
    for (uint8_t i = 0; i < length; ++i)
    {
        printf("%02d: ", i);
        for (int8_t j = 31; j >= 0; --j)
        {
            printf("%d", (buffer[i] >> j) & 0x01);
            if (j % 8 == 0)
            {
                printf(" ");
            }
        }

        printf("  |  0x%08x\n", buffer[i]);
    }
}

static void bitStitcher_helper_printBytes(uint8_t* buffer, uint8_t length)
{
    printf("BYTES\n");
    for (uint8_t i = 0; i < length; ++i)
    {
        printf("%02d: ", i);
        for (int8_t j = 7; j >= 0; --j)
        {
            printf("%d", (buffer[i] >> j) & 0x01);
        }

        printf("  |  0x%02x\n", buffer[i]);
    }
}

void print_raw(uint8_t* buffer, uint16_t size)
{
    Protocol_Frame_t* frame = (Protocol_Frame_t*)buffer;

    printf("/* TLV header: type = %u, length = %u */\n\r", frame->type, frame->length);
    printf("0x%02x, 0x%02x, 0x%02x, \n\r", buffer[0], buffer[1], buffer[2]);
    printf("/* Value buffer */\n\r");
    for (uint8_t i = TLV_HEADER_LENGTH; i < size;)
    {
        for (uint8_t j = 0; j < 8 && i < size; ++j, ++i)
        {
            printf("0x%02x, ", buffer[i]);
        }
        printf("\n\r");
    }
}

void test_tof(void)
{
    uint8_t raw[1000] = {0};

    Protocol_Frame_t* frame = (Protocol_Frame_t*)raw;
    frame->type             = Protocol_Frame_ToF;

    Protocol_ToFReport_t* value = (Protocol_ToFReport_t*)frame->value;

    value->did[0]  = 5;
    value->did[1]  = 10;
    value->did[2]  = 15;
    value->cycleId = 5000;

    /* There is only one A2A report */
    value->a2a.did[0]                = 20;
    value->a2a.did[1]                = 40;
    value->a2a.did[2]                = 60;
    value->a2a.rssi.receivedByAnchor = 5;
    value->a2a.rssi.receivedByTag    = 12;
    value->a2a.tof[0]                = 300;
    value->a2a.tof[1]                = 302;
    value->a2a.tof[2]                = 301;
    value->a2a.tof[3]                = 303;

    value->imu.hitCycleOffset = 40;
    value->imu.motionByte = 128 | 10;
    value->imu.tagImpactReportCount = 2;
    
    value->tagListsCount = 1;

    /* Fill tag lists */
    ToF_TagList_t* taglist0 = (ToF_TagList_t*)value->tagListsData;
    taglist0->numberOfToFs  = 3;
    taglist0->precision     = 0;
    taglist0->tagCount      = 2;

    /* Size Helpers */
    const size_t headerSize = sizeof(value->a2a) + sizeof(value->cycleId)
                              + sizeof(value->did) + sizeof(value->tagListsCount) + sizeof(value->imu);

    const size_t list0TagSize =
        sizeof(((ToF_Tag_t*)0)->did) + sizeof(((ToF_Tag_t*)0)->rssi)
        + (taglist0->numberOfToFs * (sizeof(((ToF_Tag_t*)0)->tof[0])));

    const size_t list0Size = sizeof(uint8_t) + sizeof(((ToF_TagList_t*)0)->tagCount)
                             + taglist0->tagCount * list0TagSize;

    /* Fill Tags */
    ToF_Tag_t* tag0 = (ToF_Tag_t*)&taglist0->tagData[0 * list0TagSize];

    tag0->did[0]                = 33;
    tag0->did[1]                = 34;
    tag0->did[2]                = 35;
    tag0->rssi.receivedByAnchor = 10;
    tag0->rssi.receivedByTag    = 12;
    tag0->tof[0]                = 200;
    tag0->tof[1]                = 201;
    tag0->tof[2]                = 202;

    ToF_Tag_t* tag1 = (ToF_Tag_t*)&taglist0->tagData[1 * list0TagSize];

    tag1->did[0]                = 33;
    tag1->did[1]                = 34;
    tag1->did[2]                = 36;
    tag1->rssi.receivedByAnchor = 11;
    tag1->rssi.receivedByTag    = 13;
    tag1->tof[0]                = 204;
    tag1->tof[1]                = 205;
    tag1->tof[2]                = 206;

    frame->length = headerSize + list0Size;

    ToF_TagImpactReport_t * impact1 = (ToF_TagImpactReport_t *) &frame->value[frame->length];
    impact1->did[0]         = 33;
    impact1->did[1]         = 34;
    impact1->did[2]         = 35;
    impact1->hitCycleOffset = 17;
    impact1->motionByte     = 27;
    frame->length += sizeof(ToF_TagImpactReport_t);

    ToF_TagImpactReport_t * impact2 = (ToF_TagImpactReport_t *) &frame->value[frame->length];
    impact2->did[0]         = 33;
    impact2->did[1]         = 34;
    impact2->did[2]         = 36;
    impact2->hitCycleOffset = 22;
    impact2->motionByte     = 50;
    frame->length += sizeof(ToF_TagImpactReport_t);

    print_raw(raw, frame->length + TLV_HEADER_LENGTH);
}

void test_sensors(void)
{
    uint8_t raw[1000] = {0};

    Protocol_Frame_t* frame = (Protocol_Frame_t*)raw;
    frame->type             = Protocol_Frame_Serial;


    Protocol_Sensors_t* value = (Protocol_Sensors_t*)frame->value;
    value->did[0]             = 10;
    value->did[1]             = 20;
    value->did[2]             = 30;
    value->cycleId            = 10000;
    value->slotId             = 50;

    sprintf(value->data, "AIRTLS");


    frame->length = sizeof(Protocol_Sensors_t) + strlen(value->data);

    print_raw(raw, frame->length + TLV_HEADER_LENGTH);
}

void test_compressedData(void)
{
    uint8_t raw[1000] = {0};

    Protocol_Frame_t* frame = (Protocol_Frame_t*)raw;
    frame->type             = Protocol_Frame_CompressedData;


    Protocol_Sensors_t* value = (Protocol_Sensors_t*)frame->value;
    value->did[0]             = 10;
    value->did[1]             = 20;
    value->did[2]             = 30;
    value->cycleId            = 10000;
    value->slotId             = 50;

    sprintf(value->data, "AIRTLS");

    for (uint8_t i = 0; i < 100; ++i)
    {
        value->data[i] = i;
    }

    frame->length = sizeof(Protocol_Sensors_t) + 100;

    print_raw(raw, frame->length + TLV_HEADER_LENGTH);
}

void test_accelerometerData(void)
{
    // clang-format off
    static const int32_t samples[] = {
        -3044,      2329,        1406,
        -3054,      2335,        1427,
        -3048,      2329,        1420,
        -3066,      2283,        1467,
        -3059,      2262,        1454,
        -3085,      2270,        1471,
        -120064,    -131064,     130979,
        131070,     131069,      124216,
        -28471,     -130818,     63966,
        -3044,      2329,        1406,
        -3054,      2335,        1427,
        -3048,      2329,        1420,
        -3066,      2283,        1467,
        -3059,      2262,        1454,
        -3085,      2270,        1471,
        -3059,      2262,        1454,
    };
    // clang-format on

    uint8_t raw[1000] = {0};

    Protocol_Frame_t* frame = (Protocol_Frame_t*)raw;
    frame->type             = Protocol_Frame_AccelerometerData;


    Protocol_Sensors_t* value = (Protocol_Sensors_t*)frame->value;
    value->did[0]             = 10;
    value->did[1]             = 20;
    value->did[2]             = 30;
    value->cycleId            = 1234;
    value->slotId             = 77;

    uint8_t accelerometerDataSize =
        bitStitcher_compress(samples, value->data, ARRAY_SIZE(samples), 18, 0);

    frame->length = sizeof(Protocol_Sensors_t) + accelerometerDataSize;

    print_raw(raw, frame->length + TLV_HEADER_LENGTH);
}

void test_gyroscopeData(void)
{
    // clang-format off
    static const int32_t samples[] = {
        -647, -229, 427,
        -333, -229, 274,
        -374, -229, 427,
        -206, -167, 290,
        -390, -188, 290,
        -229, -251, 381,
        -354, -229, 381,
        -312, -229, 274,
        -167, -206, 290,
        -251, -229, 381,
        -229, -333, 274,
        -229, -374, 427,
        -188, -390, 290,
        -229, -354, 381,
        -229, -647, 427,
        -229, -312, 274
    };
    // clang-format on

    uint8_t raw[1000] = {0};

    Protocol_Frame_t* frame = (Protocol_Frame_t*)raw;
    frame->type             = Protocol_Frame_GyroscopeData;


    Protocol_Sensors_t* value = (Protocol_Sensors_t*)frame->value;
    value->did[0]             = 10;
    value->did[1]             = 20;
    value->did[2]             = 30;
    value->cycleId            = 4321;
    value->slotId             = 77;

    uint8_t accelerometerDataSize =
        bitStitcher_compress(samples, value->data, ARRAY_SIZE(samples), 19, 0);

    frame->length = sizeof(Protocol_Sensors_t) + accelerometerDataSize;

    print_raw(raw, frame->length + TLV_HEADER_LENGTH);
}

void test_magnetometerData(void)
{
    // clang-format off
    static const int32_t samples[] = {
        -248, -579, -425,
        -259, -580, -430,
        -258, -577, -429,
        -258, -577, -432,
        -257, -579, -432,
        -248, -578, -429,
        -259, -580, -425,
        -257, -578, -432,
        -259, -577, -430,
        -258, -580, -425,
        -257, -577, -432,
        -248, -579, -429,
        -259, -578, -430,
        -258, -577, -425,
        -248, -580, -432,
        -258, -578, -429
    };
    // clang-format on

    uint8_t raw[1000] = {0};

    Protocol_Frame_t* frame = (Protocol_Frame_t*)raw;
    frame->type             = Protocol_Frame_MagnetometerData;


    Protocol_Sensors_t* value = (Protocol_Sensors_t*)frame->value;
    value->did[0]             = 10;
    value->did[1]             = 20;
    value->did[2]             = 30;
    value->cycleId            = 1111;
    value->slotId             = 33;

    uint8_t accelerometerDataSize =
        bitStitcher_compress(samples, value->data, ARRAY_SIZE(samples), 11, 0);

    frame->length = sizeof(Protocol_Sensors_t) + accelerometerDataSize;

    print_raw(raw, frame->length + TLV_HEADER_LENGTH);
}

void test_impactDataLong(void)
{
    uint8_t raw[1000] = {0};

    Protocol_Frame_t* frame = (Protocol_Frame_t*)raw;
    frame->type             = Protocol_Frame_ImpactData;


    Protocol_Sensors_t* value = (Protocol_Sensors_t*)frame->value;
    value->did[0]             = 10;
    value->did[1]             = 20;
    value->did[2]             = 30;
    value->cycleId            = 1111;
    value->slotId             = 33;

    Sensors_Motion_t * motion = value->data;
    motion->hitPosition = 17;
    motion->impact.lsb = 0x10;
    motion->impact.msb = 0x04;
    motion->length = 1;

    frame->length = sizeof(Protocol_Sensors_t) + 3u;

    print_raw(raw, frame->length + TLV_HEADER_LENGTH);
}

void test_impactDataShort(void)
{
    uint8_t raw[1000] = {0};

    Protocol_Frame_t* frame = (Protocol_Frame_t*)raw;
    frame->type             = Protocol_Frame_ImpactData;


    Protocol_Sensors_t* value = (Protocol_Sensors_t*)frame->value;
    value->did[0]             = 10;
    value->did[1]             = 20;
    value->did[2]             = 30;
    value->cycleId            = 1111;
    value->slotId             = 33;
    
    Sensors_Motion_t * motion = value->data;
    motion->hitPosition = 17;
    motion->impact.lsb = 0x04;
    motion->length = 1;

    frame->length = sizeof(Protocol_Sensors_t) + 2u;

    print_raw(raw, frame->length + TLV_HEADER_LENGTH);
}

void test_adminDistribution(void)
{
    uint8_t raw[1000] = {0};

    Protocol_Frame_t* frame = (Protocol_Frame_t*)raw;
    frame->type             = Protocol_Frame_AdminDistributionData;
    frame->length           = sizeof(Protocol_AdminDistributionInfo_t);

    Protocol_AdminDistributionInfo_t* value =
        (Protocol_AdminDistributionInfo_t*)frame->value;

    value->did[0]     = 5;
    value->did[1]     = 10;
    value->did[2]     = 15;
    value->cycleId    = 7000;
    value->apiVersion = 1;
    value->fwVersion  = 2;
    value->hwid       = 3;

    value->MAC[0] = 0x11;
    value->MAC[1] = 0x12;
    value->MAC[2] = 0x13;
    value->MAC[3] = 0x14;
    value->MAC[4] = 0x15;
    value->MAC[5] = 0x16;

    value->serialNumber[0] = 1;
    value->serialNumber[1] = 0;
    value->serialNumber[2] = 0;
    value->serialNumber[3] = 0;

    value->license[0] = 221;
    value->license[1] = 222;

    value->cycle.deviceOffset                 = 0;
    value->cycle.frequency                    = 50;
    value->cycle.seat                         = 34;
    value->bt.txPowerPreset                   = 2;
    value->uwb.ch3AntennaDelay                = 99;
    value->uwb.ch5AntennaDelay                = 100;
    value->uwb.txPowerPreset                  = 1;
    value->sensorsConfig.boardConfig          = 0x00;
    value->sensorsConfig.pressureCallibration = 10;
    value->sensorsConfig.magnetoBias.x        = 7;
    value->sensorsConfig.magnetoBias.y        = 8;
    value->sensorsConfig.magnetoBias.z        = 9;
    value->sensorsConfig.magnetoScalar.x      = 4;
    value->sensorsConfig.magnetoScalar.y      = 5;
    value->sensorsConfig.magnetoScalar.z      = 6;
    value->sensorsConfig.accelerometerBias.x  = 1;
    value->sensorsConfig.accelerometerBias.y  = 2;
    value->sensorsConfig.accelerometerBias.z  = 3;
    value->sensorsData.vbat                   = 30;
    value->sensorsData.motionByte             = 1;
    value->sensorsData.hit                    = 70;
    value->sensorsData.pressure               = 98;
    value->sensorsData.heartRate              = 99;
    value->sensorsData.mox                    = 100;

    sprintf(value->name, "AIRTLS");
    sprintf(value->group, "AB");

    print_raw(raw, frame->length + TLV_HEADER_LENGTH);
}

void test_shutdown(void)
{
    uint8_t raw[1000] = {0};

    Protocol_Frame_t* frame = (Protocol_Frame_t*)raw;
    frame->type             = Protocol_Frame_Sync_Shutdown;
    frame->length           = sizeof(SM_Protocol_CommandShutdown_t);

    SM_Protocol_CommandShutdown_t* value =
        (SM_Protocol_CommandShutdown_t*)frame->value;
    
    value->device[0].did[0] = 0x01;
    value->device[0].did[1] = 0x02;
    value->device[0].did[2] = 0x03;
    
    value->device[1].did[0] = 0x00;
    value->device[1].did[1] = 0x00;
    value->device[1].did[2] = 0x00;

    value->device[2].did[0] = 0x21;
    value->device[2].did[1] = 0x22;
    value->device[2].did[2] = 0x23;

    value->device[3].did[0] = 0x12;
    value->device[3].did[1] = 0x22;
    value->device[3].did[2] = 0x32;

    print_raw(raw, frame->length + TLV_HEADER_LENGTH);

}

void test_blink(void)
{
    uint8_t raw[1000] = {0};

    Protocol_Frame_t* frame = (Protocol_Frame_t*)raw;
    frame->type             = Protocol_Frame_Sync_Blink;
    frame->length           = sizeof(SM_Protocol_CommandBlink_t);

    SM_Protocol_CommandBlink_t* value =
        (SM_Protocol_CommandBlink_t*)frame->value;
    
    value->device[0].did[0] = 0x01;
    value->device[0].did[1] = 0x02;
    value->device[0].did[2] = 0x03;
    value->device[0].brightness = 0x03;
    value->device[0].pattern = 0x06;
    value->device[0].cycleId = 2200;
    
    value->device[1].did[0] = 0x00;
    value->device[1].did[1] = 0x00;
    value->device[1].did[2] = 0x00;
    value->device[1].brightness = 0x01;
    value->device[1].pattern = 0x00;
    value->device[1].cycleId = 2200;

    print_raw(raw, frame->length + TLV_HEADER_LENGTH);
}

void test_seatAllocation(void)
{
    uint8_t raw[1000] = {0};

    Protocol_Frame_t* frame = (Protocol_Frame_t*)raw;
    frame->type             = 254;
    frame->length           = sizeof(SM_Protocol_CommandSetSeatBasic_t);

    SM_Protocol_CommandSetSeatBasic_t* value =
        (SM_Protocol_CommandSetSeatBasic_t*)frame->value;
    
    value->device[0].did[0] = 0x01;
    value->device[0].did[1] = 0x02;
    value->device[0].did[2] = 0x03;
    value->device[0].seat = 1;
    
    value->device[1].did[0] = 0x10;
    value->device[1].did[1] = 0x20;
    value->device[1].did[2] = 0x30;
    value->device[1].seat = 5;

    value->device[2].did[0] = 0x11;
    value->device[2].did[1] = 0x22;
    value->device[2].did[2] = 0x33;
    value->device[2].seat = 10;


    print_raw(raw, frame->length + TLV_HEADER_LENGTH);
}

void test_setUWBConfig(void)
{
    uint8_t raw[1000] = {0};

    Protocol_Frame_t* frame = (Protocol_Frame_t*)raw;
    frame->type             = Protocol_Frame_SetUWBConfig;
    frame->length           = sizeof(SM_Protocol_CommandSetUWBConfig_t);

    SM_Protocol_CommandSetUWBConfig_t* value =
        (SM_Protocol_CommandSetUWBConfig_t*)frame->value;
    
    value->device[0].did[0] = 0x01;
    value->device[0].did[1] = 0x02;
    value->device[0].did[2] = 0x03;
    value->device[0].ch3 = -5;
    value->device[0].ch5 = 5;
    value->device[0].txPower = 1;
    
    value->device[1].did[0] = 0x10;
    value->device[1].did[1] = 0x20;
    value->device[1].did[2] = 0x30;
    value->device[1].ch3 = 10;
    value->device[1].ch5 = -10;
    value->device[1].txPower = 1;

    print_raw(raw, frame->length + TLV_HEADER_LENGTH);
}

int main()
{
    test_tof();
    test_sensors();
    test_adminDistribution();
    test_compressedData();
    test_accelerometerData();
    test_gyroscopeData();
    test_shutdown();
    test_blink();
    test_seatAllocation();
    test_setUWBConfig();
    
    test_impactDataLong();
    test_impactDataShort();
    test_magnetometerData();
    
    return 0;
}